// Copyright (C) 1998 Logi Ragnarsson

package org.logi.crypto.random;

import java.util.Random;
import java.io.InputStream;
import java.io.IOException;

/**
 * This class reads bits from an InputStream object and returns them
 * as random values. No randomness checking is done and an Error is
 * thrown if the end of the Reader is ever reached.
 * <p>
 * This class is useful f.ex. for reding random bits from the
 * <code>/dev/random</code> or <code>/dev/urandom</code> devices
 * where they are available (such as in linux). This would be done with
 * the following code:
 * <blockquote><pre>
 * Random rand;
 * try {
 *   rand=new RandomFromReader(new FileInputStream("/dev/random"));
 * } catch (FileNotFoundException e) {
 *   rand=new RandomSpinner();
 * }
 * </pre></blockquote>
 *
 * @version 1.0.6
 *
 * @author <a href="http://www.logi.org/~logir/">Logi Ragnarsson</a>
 *         (<a href="mailto:logir@logi.org">logir@logi.org</a>) */
public class RandomFromStream extends Random {
    
    private InputStream in;

    private static String emptyMsg = "RandomFromStream ran out of random bytes.";
    
    /**
     * Create a new RandomFromStream obejct. Random bits are read from
     * <code>in</code>
     */
    public RandomFromStream(InputStream in){
        this.in=in;
    }
    
    /** Generates a user specified number of random bytes. */
    public void nextBytes(byte[] bytes){
        int left=bytes.length;
        int i=0;
        try{
            while(left>0){
                int n = in.read(bytes,i,left);
                if(n==-1)
                    throw new Error(emptyMsg);
                left -= n;
            }
        } catch (IOException e) {
            throw new Error(emptyMsg);
        }
    }
    
    /** Generates the next random number. */
    protected synchronized int next(int bits){
        try{
            int a = in.read();
            int b = in.read();
            int c = in.read();
            int d = in.read();
            if( (a|b|c|d) < 0 )
                throw new Error(emptyMsg);
            return (a<<24) | (b<<16) | (c<< 8) | (d);
        } catch (IOException e) {
            throw new Error(emptyMsg);
        }
    }
    
}
