// Copyright (C) 1998 Logi Ragnarsson

package org.logi.crypto.test;
import org.logi.crypto.*;
import org.logi.crypto.hash.*;
import org.logi.crypto.sign.*;
import org.logi.crypto.keys.*;

import java.util.Random;
import java.io.*;

/**
 * This application tests the various Key classes.
 *
 * @author <a href="http://www.logi.org/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@logi.org">logir@logi.org</a>)
 *
 * @see org.logi.crypto.keys.Key
 */
public class TestKey extends Crypto{

    private TestKey(){}
    
    // Insecure but fast pseud-random number generator
    private static Random rand=new Random();  
    
    private static void help(){
        System.out.println("Use: java org.logi.crypto.test.TestKey <type>");
        System.out.println("where <type> is one of DES, TriDES, Blowfish");
        System.out.println("                       RSA, DH or Caesar.");
    }
    
    private static char[] metricPrefix= { 'a', 'p', 'n', 'm', ' ', 'K', 'M', 'G', 'T' };
    private static int prefixNeutral = 4;
    private static int prefixMax     = metricPrefix.length-prefixNeutral-1;
    
    private static String metricString(double x, int base){
        if (x==0)
            return "0";
        int unit = 0;
        while(x>10*base && unit<prefixMax){
            x/= base;
            unit+=1;
        }
        while(x<base/10 && unit>-prefixNeutral){
            x*=base;
            unit-=1;
        }
        String r = String.valueOf((int)(x+0.5))+" ";
        if (unit==0)
            return r;
        if (-prefixNeutral<unit && unit < prefixMax)
            return r+metricPrefix[unit+prefixNeutral];
        return r+"e"+(3*unit)+" ";
    }
    
    private static KeyPair createKeys(String[] arg){
        long start=System.currentTimeMillis();
        KeyPair kp=null;
        if(arg.length==0){
            help();
            return null;
        }
        System.out.println("KEY GENERATION");
        
        if(arg[0].equals("DES")){
            Key k = new DESKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("TriDES")){
            Key k = new TriDESKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("Blowfish")){
            Key k = new BlowfishKey();
            kp = new KeyPair(k,k);
        } else if(arg[0].equals("RSA")){
            // This is just a test, so 512 bits is enough.
            kp = RSAKey.createKeys(512);
        } else if(arg[0].equals("DH")){
            // This is just a test, so 512 bits is enough.
            DHKey k = new DHKey(512);
            kp = new KeyPair(k.getPublic(),k);
        } else if(arg[0].equals("Caesar")){
            Key k = new CaesarKey();
            kp = new KeyPair(k,k);
        } else {
            help();
            return null;
        }
        long stop=System.currentTimeMillis();

        printKeys(kp.getPublic(),kp.getPrivate());
        System.out.println("Time="+metricString(0.001*(stop-start),1000)+"s");
        
        return kp;
    }
    
    private static void printKeys(Key pub, Key pri){
        if(pri instanceof SymmetricKey){
            System.out.print(pub);
        } else {
            System.out.print(pub);
            System.out.print(pri);
        }
        
    }

    private static boolean testCDS(Key pub, Key pri){
        System.out.println("\nTESTING CONVERSION TO AND FROM STRINGS");
        boolean ok;
        try{
            String cds=pub.toString();
            Key k = (Key)fromString(cds);
            System.out.println(k);
            ok = pub.equals(k);
            if(!(pri instanceof SymmetricKey)){
                cds=pri.toString();
                k = (Key)fromString(cds);
                System.out.println(k);
                ok = ok && pri.equals(k);
            }
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }

    private static boolean testCrypt(CipherKey pub, CipherKey pri){
        System.out.println("\nTESTING ENCRYPTION/DECRYPTION");
        boolean ok=true;
        try{
            int pbs = pub.plainBlockSize();
            int cbs = pub.cipherBlockSize();
            
            int size = pub.equals(pri) ? 256*1024 : 8*1024;
            int numBlock = (size+pbs/2)/pbs;
            size = numBlock*pbs;
            
            byte[] plain  = new byte[size];
            byte[] cipher = new byte[numBlock*cbs];
            byte[] plain2 = new byte[size];
            
            java.util.Random rand = new java.util.Random();
            rand.nextBytes(plain);
            rand.nextBytes(cipher);
            
            System.out.println("Plain/cipher block sizes = "
                               +pub.plainBlockSize()+" B, "
                               +pub.cipherBlockSize()+" B");
            System.out.println("Buffer size = "+metricString(size,1024)+"B");
            
            Thread.sleep(1000); // To give the system time to settle down...
            System.out.print("Encrypting:");
            long start=System.currentTimeMillis();
            for (int i=0; i<numBlock; i++)
                pub.encrypt(plain,i*pbs, cipher,i*cbs);
            long stop=System.currentTimeMillis();
            System.out.println("\ttime="+metricString(0.001*(stop-start),1000)+
                               "s\t throughput="+metricString(1000.0*size/(stop-start),1024)+"B/s"
                              );
            
            start=System.currentTimeMillis();
            System.out.print("Decrypting:");
            for (int i=0; i<numBlock; i++)
                pri.decrypt(cipher,i*cbs, plain2,i*pbs);
            stop = System.currentTimeMillis();
            System.out.println("\ttime="+metricString(0.001*(stop-start),1000)+
                               "s\t throughput="+metricString(1000.0*size/(stop-start),1024)+"B/s"
                              );
            
            System.out.print("Checking...");
            for(int i=0; i<size; i++)
                if(plain[i] != plain2[i]){
                    System.out.println("Error at offset "+i);
                    ok=false;
                    break;
                }
            System.out.println();
            
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }

    private static boolean testVector(String algorithm){
        System.out.println("\nTESTING AGAINST TEST-VECTORS");
        boolean ok=true;
        Reader in;
        try {
            in = new FileReader("vectors."+algorithm);
        } catch (IOException e) {
            System.out.println("Unable to read vectors."+algorithm+" from the current directory.");
            System.out.println("You may need to run the test program from the test directory.");
            System.out.println("Test skipped.");
            return true;
        }
        try {
            StreamTokenizer tok = new StreamTokenizer(in);
            tok.ordinaryChars('0','9');
            tok.wordChars('0','9');
            while(tok.nextToken()==StreamTokenizer.TT_WORD){
                byte[] keyBytes = fromHexString(tok.sval);
                tok.nextToken();
                byte[] plain = fromHexString(tok.sval);
                tok.nextToken();
                byte[] cipher = fromHexString(tok.sval);
                CipherKey key = Crypto.makeSessionKey(algorithm+"Key",keyBytes);
                byte[] cipher2 = new byte[cipher.length];
                key.encrypt(plain,0, cipher2,0);
                if(!equal(cipher,cipher2)){
                    System.out.print("Failed: ");
		    ok=false;
                } else {
                    System.out.print("Passed: ");
                }
                System.out.println(key+"  "+hexString(plain)+"  "+hexString(cipher));
            }
        } catch (IOException e){
            e.printStackTrace();
            return false;
        } catch (InvalidCDSException e){
            e.printStackTrace();
            return false;
        }
        return ok;
    }
    
    private static boolean testSign(SignatureKey pub, SignatureKey pri){
        System.out.println();
        boolean ok;
        try {
            byte[] buffer = new byte[1024];
            rand.nextBytes(buffer);
            Fingerprint fp=Fingerprint.create(buffer,"SHA1");
            System.out.println("Good signature:");
            Signature s1=pri.sign(fp);
            System.out.println(s1);
            boolean m1 = pub.verify(s1,fp);
            System.out.println("Signature verified: "+(m1?"Yes":"No"));

            System.out.println();
            System.out.println("Bogus signature:");
            byte[] bogus = s1.getBytes();
            bogus[3]^=5;
            Signature s2=new Signature(bogus, "SHA1", pub.getFingerprint());
            System.out.println(s2);
            boolean m2=pub.verify(s2,fp);
            System.out.println("Signature verified: "+(m2?"Yes":"No"));

            ok = m1 && !m2 ;
        } catch (Throwable e){
            System.out.println(e);
            e.printStackTrace(System.out);
            ok = false;
        }
        System.out.println("Test "+ (ok?"passed":"failed"));
        return ok;
    }
    
    /** Run with argument DES, TriDES, Blowfish, RSA or Caesar. */
    public static void main(String[] arg) throws Exception {
        Crypto.initRandom();
        
        KeyPair kp = createKeys(arg);
        if (kp==null)
            return;
        
        Key pub = kp.getPublic();
        Key pri = kp.getPrivate();
        
        boolean ok=true;
        
        ok &= testCDS(pub,pri);
        
        if(pub instanceof CipherKey){
            ok &= testCrypt((CipherKey)pub,(CipherKey)pri);
            ok &= testVector(pub.getAlgorithm());
        }
        
        if(pub instanceof SignatureKey)
            ok &= testSign((SignatureKey)pub,(SignatureKey)pri);
        
        System.out.println();
        if(ok)
            System.out.println("All tests passed");
        else
            System.out.println("Some tests failed");
    }
    
}
